/*
 * StopList.java
 * Copyright (c) 2004, KTH NADA.
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, see
 * http://opensource.org/licenses/gpl-license.php
 *
 * Martin Hassel, 2004-Oct-06
 * http://www.nada.kth.se/~xmartin/
 */

package moj.lang;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashSet;
import java.util.Set;

import moj.util.ArrayTools;
import moj.util.FileTools;

/**
 * A <code>StopList</code> holds a set of index terms that are to be stripped 
 * away from a text (i.e. "stopped") before it is further processed. It also 
 * supports thresholds in form of minimum and maximum allowed length of index 
 * terms as well as minimum number of tokes ("words") a document must 
 * hold in order to be processed.<br />
 * <br />
 * These thresholds can be set in a properties file. The properties file can 
 * contain any of the following items:<br />
 * stoplist_file = &lt;stopword file&gt;<br />
 * shortest_word = &lt;length in characters&gt;<br />
 * longest_word = &lt;length in characters&gt;<br />
 * minimum_words_per_file = &lt;length in words&gt;<br />
 * <br />
 * The file containing the list of stop words 
 * should have one stop word per line, and stop words should be in the first 
 * column if several columns are used for different data (tab "\t" is used as 
 * column separator.
 * <br />
 * Based upon an idea by Magnus Rosell in the revised OrdHP package.
 * 
 * @author  Martin Hassel
 * @version 2006-Oct-30
 */
public class StopList {
	private HashSet<String> _stopwords;
	private SLprops _props;

	/**
	 * Creates a <code>StopList</code> using the default properties file 'StopList.properties'
	 */
	public StopList() {
		_props = new SLprops("StopList.properties");
		_stopwords = readStopList(_props.stoplistFile);
	}

	/**
	 * Creates a <code>StopList</code> using the stop list properties <code>properties</code>
	 */
	public StopList(SLprops properties) {
		_props = properties;
		_stopwords = readStopList(_props.stoplistFile);
	}

	/**
	 * Creates a <code>StopList</code> using the properties file <code>propertiesFile</code>
	 */
	public StopList(String propertiesFile) {
		_props = new SLprops(propertiesFile);
		_stopwords = readStopList(_props.stoplistFile);
	}

	/**
	 * Gets the <code>Properties</code> for the <code>StopList</code>
	 * @return the <code>Properties</code> for the <code>StopList</code>
	 */
	public SLprops getProperties() {
		return _props;
	}

	/**
	 * Gets the number of stop words contained in the StopList
	 * @return the number of stop words contained in the StopList
	 */
	@Deprecated
	public int getNrOfStopWords() {
		return _stopwords.size();
	}

	/**
	 * Sets the set of stop words to the provided <code>HashSet</code>.
	 * @param stopwords <code>HashSet</code> containing the stop words to be 
	 *        removed from a text.
	 * @return a <code>Set</code> containing the previous set of stop words.
	 */
    public Set setStopWords(HashSet<String> stopwords) {
    	Set tempsw = _stopwords;
		_stopwords = stopwords;
		return tempsw;
    }

    /**
     * Return the current <code>Set</code> of stop words to be removed from a text.
     * @return a <code>Set</code> containing the current set of stop words.
     */
	public Set getStopWords() {
		return _stopwords;
	}

	/**
	 * Add stop word to the StopLists internal list of stop words.
	 * 
	 * @param stopword stop word to add
	 */
	public void addStopWord(String stopword) {
		_stopwords.add(stopword);
	}
	
	/**
	 * Add a set of stop words from an array containing exactly one stop word 
	 * per element.
	 * 
	 * @param stopwords stop words to be added to the StopList
	 */
	public void addStopWords(String[] stopwords) {
		for(int i = 0; i < stopwords.length; i++)
			_stopwords.add(stopwords[i]);
	}

	/**
	 * Removes stop words in <code>StopList</code> from the <code>String text</code>
	 * 
	 * @param text the text which is to have stop words removed
	 * @return <code>text</code> with stop words removed 
	 */
	public String removeStopWords(String text) {
		return removeStopWords(text, false);
	}

	/**
	 * Removes stop words in <code>StopList</code> from the <code>String text</code> 
	 * and output the number of removed words to <code>System.out</code>
	 * 
	 * @param text the text which is to have stop words removed
	 * @param verbose output the number of removed words to <code>System.out</code> 
	 *        (<code>true</code>/<code>false</code>)
	 * @return <code>text</code> with stop words removed 
	 */
	public String removeStopWords(String text, boolean verbose) {
		return ArrayTools.join(removeStopWords(text.replaceAll("\\s+"," ").split("\\s"),verbose)," ");
	}

	/**
	 * Removes stop words in <code>StopList</code> from the <code>String</code> array <code>text</code>
	 * 
	 * @param text the text which is to have stop words removed
	 * @return <code>text</code> with stop words removed 
	 */
	public String[] removeStopWords(String[] text) {
		return removeStopWords(text, false);
	}

	/**
	 * Removes stop words in <code>StopList</code> from the <code>String</code> array <code>text</code>
	 * and, optionally, output the number of removed words to <code>System.out</code>
	 * 
	 * @param text the text which is to have stop words removed
	 * @param verbose output the number of removed words to <code>System.out</code> 
	 *        (<code>true</code>/<code>false</code>)
	 * @return <code>text</code> with stop words removed 
	 */
	public String[] removeStopWords(String[] text, boolean verbose) {
		int wordsRemoved = 0;
		if(_stopwords.size() > 0 || _props.shortestWord != 0
				|| _props.longestWord != 0 || _props.minimumWordsPerFile != 0) {
			boolean save;
			int wordCount = 0;
			String[] newText = new String[text.length];
			for(int i = 0; i < text.length; i++) {
				save = true;
				if(_stopwords.contains(text[i]))
					save = false;
				if(_props.shortestWord !=0 && text[i].length() <= _props.shortestWord)
					save = false;
				if(_props.longestWord !=0 && text[i].length() >= _props.longestWord)
					save = false;
				if(save) {
					newText[wordCount] = text[i];
					wordCount++;
				} else
					wordsRemoved++;
			}
			if(wordCount > _props.minimumWordsPerFile) {
				Class arrayClass = newText.getClass().getComponentType();
				String[] nv = (String[])java.lang.reflect.Array.newInstance(arrayClass, wordCount);
				System.arraycopy(newText,0,nv,0,wordCount > newText.length ? newText.length : wordCount);
				text = nv;
			}
		}
		if(verbose)
			System.out.print(wordsRemoved + " tokens stopped ");
		return text;
	}

	/**
	 * Reads the list of stop words into the <code>StopList</code>. 
	 * @param stopList filename of the file containg the list of stop words
	 * @return a <code>HashSet</code> containing the stop words
	 */
	private HashSet<String> readStopList(String stopList) {
		HashSet<String> stopSet = new HashSet<String>();
		if(_props.stoplistFile.length() > 0)
			try {
				BufferedReader fr = new BufferedReader(new FileReader(stopList));
				String word = fr.readLine();
				// If the StopList we're reading from is a FrequencyList
				if(word != null && word.indexOf("\t") > 0)
					word = word.split("\t")[0];
				while(word != null) {
					// Do not add comments
					if(!word.startsWith("#"))
						stopSet.add(word);
					word = fr.readLine();
					if(word != null && word.indexOf("\t") > 0)
						word = word.split("\t")[0];
				}
				fr.close();
			} catch(IOException ioe) {
				ioe.printStackTrace();
			}
		return stopSet;
	}

	/**
	 * Usage: moj.lang.StopList &lt;file&gt;<br />
	 * &lt;file&gt; : file to remove stopwords from<br />
	 * (&lt;properties&gt;) : properties file<br />
	 * <br />
	 * The properties file can contain any of the following items:<br />
	 * stoplist_file = &lt;stopword file&gt;<br />
	 * shortest_word = &lt;length in characters&gt;<br />
	 * longest_word = &lt;length in characters&gt;<br />
	 * minimum_words_per_file = &lt;length in words&gt;<br />
	 */
	public static void main(String[] args) {
        if(args.length == 0) {
    		System.out.println("*** StopList ***");
    		System.out.println("Usage: moj.lang.StopList <file> (<properties>)");
    		System.out.println("<file> : file to remove stopwords from");
    		System.out.println("<properties> : properties file");
    		System.exit(0);
    	}
        StopList sl;
        if(args.length > 1)
    		sl = new StopList(args[1]);
        else
    		sl = new StopList();
		GenericTokenizer gt = new GenericTokenizer();
		String tokens[] = gt.wrapText(sl.removeStopWords(FileTools.readFile2String(args[0])), 80);
		for(int token = 0; token < tokens.length; token++)
	    	System.out.println(tokens[token]);
	}
}
