/*
 * DecompounderConnection.java
 * Copyright (c) 2004, KTH NADA.
 *
 * This file is part of SweSum^2 (see http://swesum.nada.kth.se/),
 * and is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, see
 * http://opensource.org/licenses/gpl-license.php
 *
 * Martin Hassel, 2004-Feb-15
 * http://www.nada.kth.se/~xmartin/
 *
 */

package moj.lang.se;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.net.Socket;

import moj.util.FileTools;


/**
 * DecompounderConnection handles all communication with a Decompounder server.
 * Currently the decompounder only works with Swedish compounds. If you want to 
 * get all found decompounding alternatives for a compound, try using port 
 * <i>6150</i> instead of the default port <i>6140</i>. For detailed information 
 * about the compound splitting, see Sj&ouml;bergh & Kann (2004,2006).
 * 
 * @author  Martin Hassel
 * @version 2006-Oct-12
 */
public class DecompounderConnection {
	private String _host = "skrutten.nada.kth.se";
	private int _port = 6140;

    /**
     * Create a new DecompounderConnection to the given <code>host</code> 
     * at the given <code>port</code>.
     * @param host Decompounder server host to connect to.
     * @param port port on Decompounder server to connect to.
     */
    public DecompounderConnection(String host, int port) {
    	_host = host;
    	_port = port;
    }

    /**
     * Create a new DecompounderConnection to the default <code>host</code> 
     * and <code>port</code>.
     */
    public DecompounderConnection() {
    	// Uses the default host and port, as declared in header
    }

    /**
     * Splits Swedish compounds recognized by Stava into its word parts (morfs).
     * For example, "barnvagnshjul" will yield the responce "barn|vagn|s|hjul" 
     * where the word parts are separated by a vertical bar (|). The reply is 
     * given inte form of one word per line and each line has the format
     * <code>compound\t(morf|morf(|morf)*)?</code>. If Decompounder is unable to 
     * split the compound/word the second column will be empty.
     * @param text <code>String</code> containing the words that are to be 
     *        analyzed and potentially decompounded
     * @return <code>String</code> with one word per line where each line has 
     *         the format <code>compound\t(morf|morf(|morf)*)?</code>
     */
    public String decompound(String text) {
        try {
        	Socket sock = new Socket(_host, _port);
        	PrintStream output = new PrintStream(sock.getOutputStream());
        	BufferedReader input = new BufferedReader(new InputStreamReader(sock.getInputStream()));
    		StringBuffer out = new StringBuffer();
       		String line;
       		output.println(text+" ENDQ");
    		while((line = input.readLine()) != null)
    			out.append(line + " \n");
    		input.close();
    		output.close();
    		return out.toString().replaceAll("<.?reply>\\s*","");
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        return null;
    }

    /**
     * Usage: moj.lang.se.DecompounderConnection &lt;FILE|WORD|DEMO&gt; &lt;file|word&gt;<BR />
     * &lt;FILE|WORD|DEMO&gt; : decompound words in a file, given word, or print demo output<BR />
     * &lt;file|word&gt; : file to decompound the words in, or word to decompound<BR />
     */
    public static void main(String[] args) {
        if(args.length > 0) {
    		DecompounderConnection decompounder = new DecompounderConnection();
        	if(args[0].compareToIgnoreCase("DEMO") == 0) {
        		System.out.println(decompounder.decompound("barnvagnshjul husbilstak tomteverkstadsgolv"));
        	} else if(args[0].compareToIgnoreCase("FILE") == 0) {
        		System.out.println(decompounder.decompound(FileTools.readFile2String(args[1])));
        	} else if(args[0].compareToIgnoreCase("WORD") == 0) {
        		System.out.println(decompounder.decompound(args[1]));
            } else 
            	System.out.println("Unknown mode: " + args[0]);
        } else {
    		System.out.println("*** DecompounderConnection ***");
    		System.out.println("Usage: moj.lang.se.DecompounderConnection <FILE|WORD|DEMO> <file|word>");
    		System.out.println("<FILE|WORD|DEMO> : decompound words in a file, given word, or print demo output");
    		System.out.println("<file|word> : file to decompound the words in, or word to decompound");
        }
	}
}
