/*
 * GranskaConnection.java
 * Copyright (c) 2004, KTH NADA.
 *
 * This file is part of SweSum^2 (see http://swesum.nada.kth.se/),
 * and is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, see
 * http://opensource.org/licenses/gpl-license.php
 *
 * Martin Hassel, 2004-Feb-15
 * http://www.nada.kth.se/~xmartin/
 *
 */

package moj.lang.se;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.StringReader;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.Properties;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import moj.util.FileTools;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.helpers.DefaultHandler;

/**
 * GranskaConnection handles all communication with a Granska server. So far
 * tokenizing, lemmatizing, tagging, parsing and grammar checking is
 * only available for Swedish text input.
 *
 * @author  Martin Hassel
 * @version 2006-Dec-06
 */
public class GranskaConnection {
	// Default values for connecting to the Granska server at KTH NADA
	private static final String _protocol = "http";
	private static final String _host = "skrutten.nada.kth.se";
	private static final String _port = "80";
	private static final String _path = "/grimservlets/";
	private static final String _inflector_path = "inflect";
	private static final String _granska_path = "check";
	private static final String _gta_path = "analyze";
	
	private final URL _servletsBaseURL;
	/** URL to current Inflector servlet */
	public final URL inflectorURL;
	/** URL to current Granska servlet */
	public final URL granskaURL;
	/** URL to current GTA servlet */
	public final URL gtaURL;
	
	/**
	 * Create a new GranskaConnection to the Granska server given in the file 
	 * <code>Granska.properties</code>.
	 */
	public GranskaConnection() throws java.net.MalformedURLException {
		Properties props = new Properties();
		try {
			props.load(new FileInputStream("Granska.properties"));
		} catch (FileNotFoundException ex) {
			// System.err.println("Could not load properties file, will try to connect to the KTH Granska server using default values");
		} catch (IOException ex) {
			ex.printStackTrace();
		}
		_servletsBaseURL = new URL(props.getProperty("protocol",_protocol),
				props.getProperty("host",_host),
				Integer.parseInt(props.getProperty("port",_port)),
				props.getProperty("path",_path));
		inflectorURL = new URL(_servletsBaseURL, props.getProperty("inflector_path",_inflector_path));
		granskaURL = new URL(_servletsBaseURL, props.getProperty("granska_path",_granska_path));
		gtaURL = new URL(_servletsBaseURL, props.getProperty("gta_path",_gta_path));
	}
	
	/**
	 * Create a new GranskaConnection to the given <code>host</code> at the
	 * given <code>port</code> and <code>path</code>.
	 * 
	 * @param host Granska server host to connect to.
	 * @param port port on Granska server to connect to.
	 * @param path path to the servlets on the Granska server.
	 */
	public GranskaConnection(String host, int port, String path) throws java.net.MalformedURLException {
		Properties props = new Properties();
		try {
			props.load(new FileInputStream("Granska.properties"));
		} catch (FileNotFoundException ex) {
			// System.err.println("Could not load properties file, will try to connect to the KTH Granska server using default values");
		} catch (IOException ex) {
			ex.printStackTrace();
		}
		_servletsBaseURL = new URL("http", host, port, path);
		inflectorURL = new URL(_servletsBaseURL, props.getProperty("inflector_path",_inflector_path));
		granskaURL = new URL(_servletsBaseURL, props.getProperty("granska_path",_granska_path));
		gtaURL = new URL(_servletsBaseURL, props.getProperty("gta_path",_gta_path));
	}
	
	/**
	 * Create a new GranskaConnection to the Granska server given in the  
	 * <code>Properties</code> file <code>filename</code>.
	 */
	public GranskaConnection(String filename) throws java.net.MalformedURLException {
		Properties props = new Properties();
		try {
			props.load(new FileInputStream(filename));
		} catch (FileNotFoundException ex) {
			System.err.println("Could not load properties file, will try to connect to the KTH Granska server using default values");
		} catch (IOException ex) {
			ex.printStackTrace();
		}
		_servletsBaseURL = new URL(props.getProperty("protocol",_protocol),
				props.getProperty("host",_host),
				Integer.parseInt(props.getProperty("port",_port)),
				props.getProperty("path",_path));
		inflectorURL = new URL(_servletsBaseURL, props.getProperty("inflector_path",_inflector_path));
		granskaURL = new URL(_servletsBaseURL, props.getProperty("granska_path",_granska_path));
		gtaURL = new URL(_servletsBaseURL, props.getProperty("gta_path",_gta_path));
	}
	
	/**
	 * Sends the text <code>text</code> to a Granska/GTA/Inflector servlet and 
	 * returns an <code>Reader</code> "pointing" at the scrutinized text. The 
	 * returned XML-document contains information about lemmas and PoS-tags of 
	 * words as well as possible grammar errors.
	 * 
	 * @param text the text that is to be scrutinized by the Granska server.
	 * @param servlet servlet to connect to, i.e. Granska/GTA/Inflector.
	 * @return a Reader, for example to be passed to a XMLReader.
	 */
	public Reader granskaConnect(String text, URL servlet) {
		try {
			// Initiate connection to servers via the servlets
			HttpURLConnection servletConnection = (HttpURLConnection)servlet.openConnection();
			servletConnection.setDoOutput(true);
			servletConnection.setDoInput(true);
			BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(
					servletConnection.getOutputStream(), "iso-8859-1"));
			bw.write("text=" + text);
			bw.flush();
			bw.close();
			StringBuffer buf = new StringBuffer();
			BufferedReader br = new BufferedReader(new InputStreamReader(
					servletConnection.getInputStream(), "iso-8859-1"));
			// Read response from server ("XML") into StringBuffer
			for(String line = br.readLine(); line != null; line = br.readLine())
				buf.append(line);
			String xml = buf.toString();
			// The Granska/GTA servers do not output the following XML entities correctly
			// For reference, see http://www.w3.org/TR/1998/REC-xml-19980210#sec-predefined-ent
			// Order is somewhat important :)
			// quot (")
			xml = xml.replaceAll(" \" "," &#34; ");
			xml = xml.replaceAll(">\"(<|\t)",">&#34;$1");
			xml = xml.replaceAll("\"\"\"","\"&#34;\"");
			// apos (')
			xml = xml.replaceAll(" ' "," &#39; ");
			xml = xml.replaceAll(">'(<|\t)",">&#39;$1");
			xml = xml.replaceAll("\"'\"","\"&#39;\"");
			// amp (&)
			xml = xml.replaceAll(" & "," &#38; ");
			xml = xml.replaceAll(">&(<|\t)",">&#38;$1");
			xml = xml.replaceAll("\"&\"","\"&#38;\"");
			// & not followed by a ; until next < or space
			xml = xml.replaceAll("&([^;]+?)(<| )","&#38;$1$2");
			// lt (<)
			xml = xml.replaceAll(" < "," &#60; ");
			xml = xml.replaceAll("><(<|\t)",">&#60;$1");
			xml = xml.replaceAll("\"<\"","\"&#60;\"");
			// < not followed by a > until next <
			xml = xml.replaceAll("<([^>]+?)<","&#60;$1<");
			// gt (>)
			xml = xml.replaceAll(" > "," &#62; ");
			xml = xml.replaceAll(">>(<|\t)",">&#62;$1");
			xml = xml.replaceAll("\">\"","\"&#62;\"");
			// The XML is in some cases still not well formed, but should 
			// at least in most instances pass through a tolerant parser
			return new StringReader(xml);
		} catch (Exception ex) {
			ex.printStackTrace();
		}
		return null;
	}
	
	/**
	 * Sends the text <code>text</code> to a Granska server and returns an
	 * InputSource "pointing" at the scrutinized text. The returned XML-document
	 * contains information about lemmas and PoS-tags of words as well as
	 * possible grammar errors.
	 * 
	 * @param text the text that is to be scrutinized by the Granska server.
	 * @return an InputSource, for example to be passed to a XMLReader.
	 */
	public InputSource scrutinize(String text) {
		return new InputSource(granskaConnect(text, granskaURL));
	}
	
	/**
	 * Sends the text <code>text</code> to a GTA server and returns an
	 * InputSource "pointing" at the parsed text. The returned XML-document
	 * contains information about lemmas and PoS-tags of words as well as
	 * phrase structure information.
	 * 
	 * @param text the text that is to be parsed by the GTA server.
	 * @return an InputSource, for example to be passed to a XMLReader.
	 */
	public InputSource parse(String text) {
		return new InputSource(granskaConnect(text, gtaURL));
	}
	
	/**
	 * Tokenizes the given text. The tokens (words, delimiters
	 * etc.) in the returned string are separated by space. For example, the text
	 * "Han springer fortare!" would yield the response "Han springer fortare ! ".
	 * Note: Processes the text until first encountered newline, following text is skipped!
	 * 
	 * @param text the text that is to be tokenized.
	 * @return the text tokenized, or <code>null</code> if <code>tokenize</code>
	 *         fails.
	 */
	public String tokenize(String text) {
		try {
			SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
			TokenizeHandler handler = new TokenizeHandler();
			parser.parse(this.parse(text), handler);
			return handler.text;
		} catch (Exception ex) {
			ex.printStackTrace();
		}
		return null;
	}
	
	/**
	 * Tokenizes and lemmatizes the given text. The tokens (words, delimiters
	 * etc.) in the returned string are separated by space. For example, the text
	 * "Han springer fortare!" would yield the response "han springa fort ! ".
	 * Note: Processes the text until first encountered newline, following text is skipped!
	 * 
	 * @param text the text that is to be tokenized and lemmatized.
	 * @return the text tokenized and lemmatized, or <code>null</code> if
	 *         <code>lemmatize</code> fails.
	 */
	public String lemmatize(String text) {
		try {
			SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
			LemmatizeHandler handler = new LemmatizeHandler();
			parser.parse(this.parse(text), handler);
			return handler.text;
		} catch (Exception ex) {
			//ex.printStackTrace();
		}
		return null;
	}
	
	/**
	 * Tokenizes, lemmatizes and PoS-tags the given text. The tokens (words,
	 * delimiters etc.) in the returned string are separated by space. For
	 * example, the text "Han springer fortare!" would yield the response
	 * "han_pn springa_vb fort_ab !_mad ".
	 * Note: Processes the text until first encountered newline, following text is skipped!
	 * 
	 * @param text the text that is to be tokenized, lemmatized and PoS-tagged.
	 * @return the text tokenized, lemmatized and PoS-tagged - or <code>null</code>
	 *         if <code>lemmaTag</code> fails.
	 */
	public String lemmaTag(String text) {
		try {
			SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
			LemmaTagHandler handler = new LemmaTagHandler();
			parser.parse(this.parse(text), handler);
			return handler.text;
		} catch (Exception ex) {
			//ex.printStackTrace();
		}
		return null;
	}
	
	/**
	 * Deprecated: Use posTag() instead, it offers the same functionality.
	 * This method may well be completely removed in future versions.
	 */
	@Deprecated
	public String simpleTag(String text) {
		return posTag(text);
	}

	/**
	 * Tags the given text with morphosyntactic tags. The tokens (words, delimiters etc.) in the
	 * returned string are separated by space. For example, the text "Han
	 * springer fortare!" would yield the response "Han_pn.utr.sin.def.sub
	 * springer_vb.prs.akt fortare_ab.kom !_mad ".
	 * Note: Processes the text until first encountered newline, following text is skipped!
	 * 
	 * @param text the text that is to be morphosyntacticly tagged.
	 * @return the text morphosyntacticly tagged - or <code>null</code> if
	 *         <code>posTag</code> fails.
	 */
	public String posTag(String text) {
		try {
			SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
			PoSTagHandler handler = new PoSTagHandler();
			parser.parse(this.parse(text), handler);
			return handler.text;
		} catch (Exception ex) {
			ex.printStackTrace();
		}
		return null;
	}
	
	/**
	 * Partial shallow parsing of the given text. The format of the returned 
	 * string is one token (word, delimiter etc.) per row with each form of 
	 * analyzis divided by tab (\t), e.g.<br /> 
	 * "H&auml;ndelsen       nn.utr.sin.def.nom      NPB     CLB".
	 * 
	 * @param text the text that is to be parsed
	 * @return the text parsed - or <code>null</code> if
	 *         <code>parseIOB</code> fails.
	 */
	public String parseIOB(String text) {
		try {
			SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
			ParseIOBHandler handler = new ParseIOBHandler();
			parser.parse(this.parse(text), handler);
			return handler.text;
		} catch (Exception ex) {
			ex.printStackTrace();
		}
		return null;
	}
	
	/**
	 * Inflects the given word.
	 * 
	 * @param word the word that is to be inflected.
	 * @return all inflections of the given word, no matter wordclass.
	 *         Inflections are separated by space.
	 */
	public String inflect(String word) {
		return inflect(word, "");
	}
	
	/**
	 * Inflects the given word according to the paradigm of the given wordclass.
	 * 
	 * @param word the word that is to be inflected.
	 * @param wordclass the wordclass of the word that is to be inflected.
	 * @return all inflections of the given word according to the paradigm of
	 *         the given wordclass. Inflections are separated by space.
	 */
	public String inflect(String word, String wordclass) {
		try {
			InputSource source = 
				new InputSource(granskaConnect(word+"&wc="+wordclass, inflectorURL));
			SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
			InflectHandler handler = new InflectHandler();
			parser.parse(source, handler);
			return handler.text;
		} catch (Exception ex) {
			ex.printStackTrace();
		}
		return null;
	}
	
	
	// Nested ContentHandlers that extracts lemmas as the XML is parsed.
	// The result is a analyzed text in the public String "text".

	private class TokenizeHandler extends DefaultHandler {
		private StringBuffer outbuffer = new StringBuffer();
		private StringBuffer buffer = null;
		private boolean inModify = false;
		private boolean inWord = false;
		public String text = "";
		
		public void startElement(String namespaceURI, String localName,
				String qualifiedName, Attributes atts) {
			if(qualifiedName.equals("modify"))
				inModify = true;
			if(qualifiedName.equals("w"))
				inWord = true;
			buffer = new StringBuffer();
		}
		
		public void endElement(String namespaceURI, String localName,
				String qualifiedName) {
			if(qualifiedName.equals("modify"))
				inModify = false;
			if(qualifiedName.equals("w") && inModify == false) {
				outbuffer.append(buffer + " ");
				inWord = false;
			}
		}
		
		public void characters(char[] ch, int start, int length) {
			if(inWord)
				if(buffer != null)
					buffer.append(ch, start, length);
		}
		
		public void endDocument() {
			text = outbuffer.toString();
		}
	} // End nested class
	
	private class LemmatizeHandler extends DefaultHandler {
		private StringBuffer outbuffer = new StringBuffer();
		private boolean inModify = false;
		public String text = "";
		
		public void startElement(String namespaceURI, String localName,
				String qualifiedName, Attributes atts) {
			if(qualifiedName.equals("modify"))
				inModify = true;
			if(qualifiedName.equals("w") && inModify == false) {
				if(atts.getValue("lemma") != null)
					outbuffer.append(atts.getValue("lemma") + " ");
			}
		}
		
		public void endElement(String namespaceURI, String localName,
				String qualifiedName) {
			if(qualifiedName.equals("modify"))
				inModify = false;
		}

		public void endDocument() {
			text = outbuffer.toString();
		}
	} // End nested class
	
	private class LemmaTagHandler extends DefaultHandler {
		private StringBuffer outbuffer = new StringBuffer();
		private boolean inModify = false;
		public String text = "";
		
		public void startElement(String namespaceURI, String localName,
				String qualifiedName, Attributes atts) {
			if(qualifiedName.equals("modify"))
				inModify = true;
			if(qualifiedName.equals("w") && inModify == false) {
				if(atts.getValue("lemma") != null) {
					outbuffer.append(atts.getValue("lemma").replaceAll(" ","_") + "_");
					String tag = atts.getValue("tag");
					if(tag.indexOf(".") > 1)
						outbuffer.append(tag.substring(0, tag.indexOf(".")) + " ");
					else
						outbuffer.append(tag + " ");
				}
			}
		}
		
		public void endElement(String namespaceURI, String localName,
				String qualifiedName) {
			if(qualifiedName.equals("modify"))
				inModify = false;
		}

		public void endDocument() {
			text = outbuffer.toString();
		}
	} // End nested class
	
	private class PoSTagHandler extends DefaultHandler {
		private StringBuffer outbuffer = new StringBuffer();
		private StringBuffer buffer = null;
		private boolean inModify = false; 
		private boolean inWord = false; 
		private String _word = "";
		private String _tag = "";
		public String text = "";
		
		public void startElement(String namespaceURI, String localName,
				String qualifiedName, Attributes atts) {
			if(qualifiedName.equals("modify"))
				inModify = true;
			if(qualifiedName.equals("w")) {
				_tag = atts.getValue("tag");
				inWord = true;
			}
			buffer = new StringBuffer();
		}
		
		public void endElement(String namespaceURI, String localName,
				String qualifiedName) {
			if(qualifiedName.equals("modify"))
				inModify = false;
			if(qualifiedName.equals("w") && inModify == false) {
				_word = buffer.toString().replaceAll(" ","_");
				outbuffer.append(_word + "_" + _tag + " ");
				inWord = false;
			}
		}
		
		public void characters(char[] ch, int start, int length) {
			if(inWord)
				if(buffer != null)
					buffer.append(ch, start, length);
		}
		
		public void endDocument() {
			text = outbuffer.toString();
		}
	} // End nested class
	
	private class ParseIOBHandler extends DefaultHandler {
		private StringBuffer outbuffer = new StringBuffer();
		private StringBuffer buffer = null;
		private boolean inRow = false; 
		public String text = "";
		
		public void startElement(String namespaceURI, String localName,
				String qualifiedName, Attributes atts) {
			if(qualifiedName.equals("row")) {
				inRow = true;
			}
			buffer = new StringBuffer();
		}
		
		public void endElement(String namespaceURI, String localName,
				String qualifiedName) {
			if(qualifiedName.equals("row")) {
				if(buffer.length() > 0)
					outbuffer.append(buffer.toString().replaceAll(" ","_") + "\n");
				inRow = false;
			}
		}
		
		public void characters(char[] ch, int start, int length) {
			if(inRow)
				if(buffer != null)
					buffer.append(ch, start, length);
		}
		
		public void endDocument() {
			text = outbuffer.toString();
		}
	} // End nested class
	
	private class InflectHandler extends DefaultHandler {
		private StringBuffer outbuffer = new StringBuffer();
		private StringBuffer buffer = null;
		private boolean inProposal = false;
		public String text = "";
		
		public void startElement(String namespaceURI, String localName,
				String qualifiedName, Attributes atts) {
			if(qualifiedName.equals("proposal"))
				inProposal = true;
			buffer = new StringBuffer();
		}
		
		public void endElement(String namespaceURI, String localName,
				String qualifiedName) {
			if(inProposal) {
				if(qualifiedName.equals("word"))
					outbuffer.append(buffer + " ");
				if(qualifiedName.equals("proposal"))
					inProposal = false;
			}
		}
		
		public void characters(char[] ch, int start, int length) {
			if(inProposal)
				if(buffer != null)
					buffer.append(ch, start, length);
		}
		
		public void endDocument() {
			text = outbuffer.toString();
		}
	} // End nested class(es)


	/**
	 * Usage: moj.lang.se.GranskaConnection &lt;TOKENIZE|LEMMATIZE|LEMMATAG|POSTAG|PARSE|INFLECT|DEMO&gt; &lt;file|word&gt; (&lt;word class&gt;)<BR />
	 * &lt;TOKENIZE|LEMMATIZE|LEMMATAG|POSTAG|PARSE|INFLECT|DEMO&gt; : keywords denoting desired function/output, or demo output<BR />
	 * &lt;file|word&gt; : file to tag words in, or word to inflect<BR />
	 * &lt;word class&gt; : inflection paradigm (if left out forms are generated for all PoS)<BR />
	 */
	public static void main (String[] args) {
        if(args.length > 0) {
    		try {
    			GranskaConnection granska = new GranskaConnection();
            	if(args[0].compareToIgnoreCase("DEMO") == 0) {
					// Unicode escapes used for certain Swedish letters
        			System.out.println("Text: \"Textklustring f\u00f6r ihop en m\u00e4ngd texter i grupper s\u00e5 att texterna i dessa liknar varandra.\"");
        			System.out.println(granska.tokenize("Textklustring f\u00f6r ihop en m\u00e4ngd texter i grupper s\u00e5 att texterna i dessa liknar varandra."));
        			System.out.println(granska.lemmatize("Textklustring f\u00f6r ihop en m\u00e4ngd texter i grupper s\u00e5 att texterna i dessa liknar varandra."));
        			System.out.println(granska.lemmaTag("Textklustring f\u00f6r ihop en m\u00e4ngd texter i grupper s\u00e5 att texterna i dessa liknar varandra."));
        			System.out.println(granska.posTag("Textklustring f\u00f6r ihop en m\u00e4ngd texter i grupper s\u00e5 att texterna i dessa liknar varandra."));
        			System.out.println(granska.inflect("springa","nn"));
            	} else if(args[0].compareToIgnoreCase("INFLECT") == 0) {
        			String arg1 = "", arg2 = "";
                    if(args.length > 1) arg1 = args[1];
                    if(args.length > 2) arg2 = args[2];
            		System.out.println(granska.inflect(arg1, arg2));
            	} else {
            		String text = "";
            		String file[] = null;
                    if(args.length > 1)
                    	text = FileTools.readFile2String(args[1]).replaceAll("\n", " ");
                    if(args[0].compareToIgnoreCase("TOKENIZE") == 0) {
                    	file = granska.tokenize(text).split("\\s");
                    	for(int i = 0; i < file.length; i++)
                        	System.out.println(file[i]);
                    } else if(args[0].compareToIgnoreCase("LEMMATIZE") == 0) {
                    	file = granska.lemmatize(text).split("\\s");
                    	for(int i = 0; i < file.length; i++)
                        	System.out.println(file[i]);
                    } else if(args[0].compareToIgnoreCase("LEMMATAG") == 0) {
                    	file = granska.lemmaTag(text).split("\\s");
                    	if(file.length > 0)
                        	for(int i = 0; i < file.length; i++)
                            	System.out.println(file[i].substring(0,file[i].lastIndexOf("_")) +
                            			"\t" + file[i].substring(file[i].lastIndexOf("_")+1));
                    } else if(args[0].compareToIgnoreCase("POSTAG") == 0) {
                    	file = granska.posTag(text).split("\\s");
                    	if(file.length > 0)
                        	for(int i = 0; i < file.length; i++)
                            	System.out.println(file[i].substring(0,file[i].lastIndexOf("_")) + 
                            			"\t" + file[i].substring(file[i].lastIndexOf("_")+1));
                	} else if(args[0].compareToIgnoreCase("PARSE") == 0) {
                		System.out.println(granska.parseIOB(text));
                    } else 
                    	System.out.println("Unknown mode: " + args[0]);
            	}
    		} catch (Exception ex) {
    			ex.printStackTrace();
    		}
        } else {
    		System.out.println("*** GranskaConnection ***");
    		System.out.println("Usage: moj.lang.se.GranskaConnection <TOKENIZE|LEMMATIZE|LEMMATAG|POSTAG|PARSE|INFLECT|DEMO> <file|word> (<word class>)");
    		System.out.println("<TOKENIZE|LEMMATIZE|LEMMATAG|POSTAG|PARSE|INFLECT|DEMO> : keywords denoting desired function/output, or demo output");
    		System.out.println("<file|word> : file to tag words in, or word to inflect");
    		System.out.println("<word class> : inflection paradigm (if left out forms are generated for all PoS)");
        }
	}
}
