package moj.lang.statserv;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.StringReader;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

/**
 * StatServConnection handles all communication with a StatServ server.
 *
 * @author  Martin Hassel
 * @version 2006-Dec-13
 */
public class StatServConnection {
	// Default values for connecting to the Granska server at KTH NADA
	private static final String _protocol = "http";
	private static final String _host = "ordbok.nada.kth.se";
	private static final String _port = "8082";
	private static final String _path = "/statserv/";
	private static final String _statserv_path = "FrontController";
	/** URL to current StatServ servlet **/
	public final URL statservURL;
	// Modules to be used in the request
	private final List<StatServRequest> modules = new ArrayList<StatServRequest>();

	/**
	 * Create a new StatServConnection to the StatServ server given in the file
	 * <code>StatServ.properties</code>.
	 */
	public StatServConnection() throws java.net.MalformedURLException {
		Properties props = new Properties();
		try {
			props.load(new FileInputStream("StatServ.properties"));
		} catch(FileNotFoundException ex) {
			// System.err.println("Could not load properties file, will try to connect to the KTH StatServ server using default values");
		} catch(IOException ex) {
			ex.printStackTrace();
		}
		statservURL = new URL(new URL(props.getProperty("protocol",_protocol),
				props.getProperty("host",_host),
				Integer.parseInt(props.getProperty("port",_port)),
				props.getProperty("path",_path)),
				props.getProperty("statserv_path",_statserv_path));
	}

	/**
	 * Adds a <code>StatServRequest</code> module to the connection.
	 * @param module <code>StatServRequest</code> module to add
	 * @return <code>true</code> on success, else <code>false</code>
	 */
	public boolean addModule(StatServRequest module) {
		return modules.add(module);
	}

	/**
	 * Removes a <code>StatServRequest</code> module from the connection.
	 * @param module <code>StatServRequest</code> module to remove
	 * @return <code>true</code> on success, else <code>false</code>
	 */
	public boolean removeModule(StatServRequest module) {
		return modules.remove(module);
	}

	/**
	 * Removes <i>all</i> <code>StatServRequest</code> modules from the connection.
	 */
	public void removeModules() {
		modules.clear();
	}

	/**
	 * Returns the XML that constitutes a <code>StatServConnection</code> request.  
	 * @param title the title of the document statistics is requested for
	 * @param text the text statistics is requested for
	 * @return XML representing a <code>StatServConnection</code> request
	 */
	public String requestXML(String title, String text) {
		String request = "<?xml version=\"1.0\" encoding=\""
			+ java.nio.charset.Charset.defaultCharset().name()
			+ "\"?>\n"
			+ "<stat-serv-request>\n"
			+ "<document>\n<title>" + title + "</title>\n<content>"
			+ "<![CDATA[" + text.replace("\n"," ") + "]]>"
			+ "</content>\n</document>\n"
			+ "<actions>\n"
			+ (modules.get(0).localstats ? "<get-local-stats/>\n" : "")
			+ (modules.get(0).referencestats ? "<get-reference-stats>\n"
			+ "<doc-set>" + modules.get(0).docset + "</doc-set>\n"
			+ "</get-reference-stats>\n" : "")
			+ "</actions>\n";
		for(StatServRequest module : modules)
			request += module.requestXML();
		return request + "</stat-serv-request>\n";
	}

	/**
	 * Returns the XML that constitutes a <code>StatServConnection</code> answer.  
	 * @param isr the <code>InputStreamReader</code> to read the XML from
	 * @return XML representing a <code>StatServConnection</code> answer
	 */
	private String answerXML(InputStreamReader isr) {
		StringBuilder sb = new StringBuilder();
		try {
			BufferedReader br = new BufferedReader(isr);
			String line = null;
			do {
				if((line = br.readLine()) != null)
					sb.append(line+"\n");
			} while(line != null);
		} catch(Exception ex) {
			ex.printStackTrace();
		}
		return sb.toString();
	}

	/**
	 * Sends the <code>text</code> and <code>title</code> to a StatServ servlet 
	 * and lets each previously added<code>StatServRequest</code> module request 
	 * and parse its own statistics, which is then made available through each 
	 * <code>StatServRequest</code> module respectively.
	 * @param title the title of the document statistics is requested for
	 * @param text the text statistics is requested for
	 * @return <code>true</code> on success, else <code>false</code>
	 */
	public boolean StatServConnect(String title, String text) {
		try {
			// Initiate connection to servers via servlet
			HttpURLConnection servletConnection = (HttpURLConnection)statservURL.openConnection();
			servletConnection.setDoOutput(true);
			servletConnection.setDoInput(true);
			servletConnection.setRequestMethod("POST");
			servletConnection.setAllowUserInteraction(false);
			servletConnection.setUseCaches(false);
			PrintWriter pw = new PrintWriter(new OutputStreamWriter(
				servletConnection.getOutputStream(), java.nio.charset.Charset.defaultCharset().name()));
			pw.println(requestXML(title, text));
			pw.flush();
			pw.close();
			String answer = answerXML(new InputStreamReader(
				servletConnection.getInputStream(), "iso-8859-1"));
			for(StatServRequest module : modules)
				module.parse(new StringReader(answer));
//			System.out.println(requestXML(title, text));
//			System.out.println(answer);
		} catch(Exception ex) {
			//ex.printStackTrace();
			return false;
		}
		return true;
	}

	public static void main(String[] args) throws Exception {
		String title = "testtext";
		String text = "Det h\u00e4r \u00e4r ett exempel p\u00e5 hur man anv\u00e4nder sig av flera moduler som anv\u00e4nder sig av \n\u00e5 och \u00e4 och \u00f6 f\u00f6r att testa just flera moduler och \u00e5 och \u00e4 och \u00f6.";
		StatServConnection ssc = new StatServConnection();
		LemmaTagRequest ltr = new LemmaTagRequest("SVD", true, true, false);
		WordTagRequest wtr = new WordTagRequest("SVD", true, true, false);
		ssc.addModule(ltr);
		ssc.addModule(wtr);
		ssc.StatServConnect(title, text);
		// Print local statistics via lemma module
		System.out.println("Tokens in the text: " + ltr.localTokens);
		for(Iterator<String> iterator = ltr.getLocalTF().keySet().iterator(); iterator.hasNext();) {
			String lemmaTag = iterator.next();
			System.out.println(lemmaTag + "\ttf: " + ltr.getLocalTF().get(lemmaTag));
		}
		System.out.println();
		// Print reference statistics via lemma module
		System.out.println("Tokens in the data set: " + ltr.referenceTokens);
		System.out.println("Documents in the data set: " + ltr.referenceDocuments);
		for(Iterator<String> iterator = ltr.getReferenceTF().keySet().iterator(); iterator.hasNext();) {
			String lemmaTag = iterator.next();
			System.out.println(lemmaTag 
					+ "\ttf: " + ltr.getReferenceTF().get(lemmaTag)
					+ "\tdf: " + ltr.getReferenceDF().get(lemmaTag));
		}
		System.out.println();
		// Print local statistics via word module
		System.out.println("Tokens in the text: " + wtr.localTokens);
		for(Iterator<String> iterator = wtr.getLocalTF().keySet().iterator(); iterator.hasNext();) {
			String wordTag = iterator.next();
			System.out.println(wordTag + "\ttf: " + wtr.getLocalTF().get(wordTag));
		}
		System.out.println();
		// Print reference statistics via word module
		System.out.println("Tokens in the data set: " + wtr.referenceTokens);
		System.out.println("Documents in the data set: " + wtr.referenceDocuments);
		for(Iterator<String> iterator = wtr.getReferenceTF().keySet().iterator(); iterator.hasNext();) {
			String wordTag = iterator.next();
			System.out.println(wordTag 
					+ "\ttf: " + wtr.getReferenceTF().get(wordTag)
					+ "\tdf: " + wtr.getReferenceDF().get(wordTag));
		}
	}
}
