/*
 * RandomLabel.java
 * Copyright (c) 2004, KTH NADA.
 *
 * This file is part of SweSum^2 (see http://swesum.nada.kth.se/),
 * and is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, see
 * http://opensource.org/licenses/gpl-license.php
 *
 * Martin Hassel, 2004-Feb-01
 * http://www.nada.kth.se/~xmartin/
 *
 */

package moj.ri;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Random;

import moj.ri.weighting.WeightingScheme;
import moj.similarity.SimilarityMeasure;
import moj.similarity.measures.CosineD_row;
import moj.util.ArrayTools;

/**
 * A RandomLabel consists of a word (or term), a term frequency count,
 * a document frequency count, a randomly initalized label and a
 * contextually updated (weighted window) context vector.
 * This container is used by RandomIndex to store the index terms and
 * their corresponding frequency and context data.
 *
 * @author  Martin Hassel
 * @version 2006-Apr-06
 */
public class RandomLabel implements Comparable, Serializable {
    private static final long serialVersionUID = -3048832246809911566L;
    private boolean _validState = true; // Does the RL have a valid state (ie not updating)?
	private final String _word; // The word the 'label' is to be attached to
    private String _meta_data = null; // Field available for meta data (e.g. pos)
    private long _termFrequency = 0; // Number of updates done to the label
    private int  _docFrequency = 0;  // Number of documents the word occurs in
    private int _dimensionality;	// The dimensionality of the random label
    private int _negs[];	// The indexes for -1:s in the random label
    private int _poss[]; 	// The indexes for 1:s in the random label
    private float _wordContext[];	// The actual contextually updated label (word level)
    //private float _docContext[];	// The actual contextually updated label (doc level, not implemented yet)

    /**
     * To construct a new RandomLabel object we need the word that is to be 'labeled',
     * the length of the label (i.e. the label's dimensionality), the randomness of 
     * the label's initial state (i.e. number of non-zero elements) and a seed.
     * @param word the word which the label is to be associated to.
     * @param dimensionality the length of the label to be associated to the word.
     * @param randomDegree the 'degree' of randomness initially applied to the label
     *        given in total number of non-zeros in the inital label (even number).
     *        Should not be greater than dimensionalty and therefore, in this case,
     *        defualts to dimensionality.
     * @param seed a seed for the local random generator. This seed, in combination
     *        with <code>word</code>, makes it very likely that the created random label is
     *        "unique" yet reproducable.
     */
    public RandomLabel(String word, int dimensionality, int randomDegree, int seed) {
    	_validState = false;
        _word = word;
        _dimensionality = dimensionality;
        _wordContext = new float[dimensionality];

        // Nr of initial pos & neg states can never be more than the nr of elements in label
        if(dimensionality < randomDegree)
            randomDegree = dimensionality;

        // We want an equal number of initial +1 and -1 in the label integer array
        int no_initial_pos_and_neg = Math.round(randomDegree/2);
    	_negs = new int[no_initial_pos_and_neg];
    	_poss = new int[no_initial_pos_and_neg];

        // Create reproducable random generator using the word (index term) as seed
        try {
            long wseed = 0;
            byte[] wordar = word.getBytes("UTF-8");
            for(int i = 0; i < wordar.length; i++)
            	// DOG and GOD are not the same ;-)
                wseed += (wordar[i]*i);
            Random randgen = new Random(wseed+seed);

            // Generate the positions in the random label to be initially filled
            HashMap<Long,String> pos_and_neg_positions = new HashMap<Long,String>();
            for(int current = 0; current < no_initial_pos_and_neg; current++) {
                // Generate unused position and add a +1
                long position = randgen.nextInt(dimensionality);
                while(pos_and_neg_positions.get(position) != null) {
                    position = randgen.nextInt(dimensionality);
                }
                pos_and_neg_positions.put(position, "+");
                _poss[current]=(int)position;
                // Generate unused position and add a -1
                position = randgen.nextInt(dimensionality);
                while(pos_and_neg_positions.get(position) != null) {
                    position = randgen.nextInt(dimensionality);
                }
                pos_and_neg_positions.put(position, "-");
                _negs[current]=(int)position;
            }
        } catch (java.io.UnsupportedEncodingException ex) {
            ex.printStackTrace();
        }
    	_validState = true;
    }

    /**
     * Defaults the dimensionality to 1800 and the randomness to 8, i.e. four
     * +1 and four -1. Seed is set to 123.
     * @param word the word which the label is to be associated to.
     */
    public RandomLabel(String word) {
        this(word, 1800, 8, 123);
    }

    /**
     * Creates an empty RandomLabel with a dimensionality of 0 (zero). 
     * Usefull for padding.
     */
    public RandomLabel() {
        this("", 0, 0, 0);
    }

    /**
     * Construct new RandomLabel from existing data.
     * @param word the word which the label is to be associated to.
     * @param termFrequency the term frequency for the associated word.
     * @param docFrequency the document frequency for the associated word.
     * @param positivePositions the positions in the RandomLabel that should hold "1"
     * @param negativePositions the positions in the RandomLabel that should hold "-1"
     * @param context the context vector to be associated to the word
     *        (i.e. an array representing the co-occurrence "coloring").
     */
    public RandomLabel(String word, String metadata, long termFrequency, int docFrequency, 
    		int positivePositions[], int negativePositions[], float context[]) {
    	_validState = false;
        _word = word;
        _meta_data = metadata;
        _termFrequency = termFrequency;
        _docFrequency = docFrequency;
        _dimensionality = context.length;
        _poss = positivePositions;
        _negs = negativePositions;
        _wordContext = context;
    	_validState = true;
    }

    /**
     * Get the word which the RandomLabel is associated to.
     * @return the word which the RandomLabel is associated to.
     */
    public String getWord() {
        return _word;
    }

    /**
     * Get <code>String</code> based meta data for the RandomLabel. This can 
     * for example be used to store part of speach tags or distributional 
     * features.
     * @return <code>String</code> based meta data for the RandomLabel.
     */
    public String getMetaData() {
        return _meta_data;
    }

    /**
     * Set <code>String</code> based meta data for the RandomLabel. This can 
     * for example be used to store part of speach tags or distributional 
     * features.
     * @return <code>String</code> based meta data for the RandomLabel.
     */
    public String setMetaData(String metadata) {
    	_validState = false;
    	String _old_meta_data = _meta_data;
    	_meta_data = metadata;
    	_validState = true;
        return _old_meta_data;
    }

    /**
     * Get the negative positions in the random label associated to the word.
     * @return the negative part of the randomly generated label associated to the word.
     */
    public int[] getNegativePositions() {
        return _negs;
    }

    /**
     * Get the positive positions in the random label associated to the word.
     * @return the positive part of the randomly generated label associated to the word.
     */
    public int[] getPositivePositions() {
        return _poss;
    }

    /**
     * Get the context vector associated to the word.
     * @return the contextually updated (weighted window) context vector associated to
     * the word, or <code>null</code> if the label has been pruned.
     */
    public float[] getContext() {
        return _wordContext;
    }

    private float[] setContext(float[] context) {
    	float[] _old_context = _wordContext;
    	_wordContext = context;
        return _old_context;
    }

	/**
	 * Prunes the RandomLabel by setting the context vector to <code>null</code>.
	 */
	public void prune() {
		_wordContext = null;
	}

    /**
     * Get the random label associated to the word.
     * @return the randomly generated label associated to the word.
     */
    public int getDimensionality() {
        return _dimensionality;
    }

    /**
     * Get the term frequency for the current word.
     * @return the term frequency, i.e. nr of contextual updates (including initialization)
     * the RandomLabel has gone through.
     */
    public long getTermFrequency() {
        return _termFrequency;
    }

    /**
     * Increment the term frequency for the current word.
     * @return the term frequency, i.e. nr of contextual updates (including initialization)
     * the RandomLabel has gone through.
     */
    public long incrementTermFrequency() {
        return ++_termFrequency;
    }

    /**
     * Get the document frequency for the current word.
     * @return the document frequency, i.e. nr of documents the word associated with this
     * RandomLabel occurs in.
     */
    public int getDocumentFrequency() {
        return _docFrequency;
    }

    /**
     * Increment the document frequency for the current word.
     * @return the document frequency, i.e. the number of documents the word associated
     * with this RandomLabel occurs in <i>after</i> the increment.
     */
    public int incrementDocumentFrequency() {
        return ++_docFrequency;
    }

    /**
     * Tells whether the RandomLabel is in a valid state or not. The RL is typically 
     * in an invalid state during initialization and when updating the context vector.
     * @return <code>true</code> if in a valid state, otherwise <code>false</code>
     */
    public boolean validState() {
        return _validState;
    }

    /**
     * Update this RandomLabel's context with the weighted labels of the RandomLabels in left
     * and right context using the supplied weighting scheme (as defined by a visiting object
     * <code>weightingScheme</code>). The RandomLabel of the word nearest to this is the first 
     * element in the context vectors, (context window) the second nearest the second, and so 
     * on (note: this applies to <i>both</i> <code>rightContext</code> <i>and</i> 
     * <code>leftContext</code>).
     * 
     * All RandomLabels in <code>leftContext</code> and <code>rightContext</code> 
     * <i>must</i> have the same dimensionality. However, <code>leftContext</code> and 
     * <code>rightContext</code> themselves (i.e. the context window) do not have to be of 
     * equal length (i.e. you can have an unbalanced context window).
     * 
     * @param  leftContext an array of RandomLables where the first element represents the word
     *         closest to the left of the word who's label is being updated, the second element
     *         represents the word second closest to the left and so on.
     *         No slot may be empty (null) as this will cause a NullPointerException.
     * @param  rightContext same as for <code>leftContext</code> but for the right side.
     * @param  weightingScheme a visiting object that contains the methods for calculating the
     *         weights for the left resp. right contexts based upon distance to the current label.
     * @return returns true upon success, else false. The most common reason for failure is that
     *         not all RandomLabels have the same dimesionality.
     */
    public boolean updateContext(RandomLabel[] leftContext, RandomLabel[] rightContext, WeightingScheme weightingScheme) {
    	_validState = false;
    	boolean retval = true;
        // If the context vector is set to null the label itself will never be
        // updated, but it will still contextually update other random labels
        if(_wordContext != null) {
            RandomLabel tempRL = this;
            // Update focusContext with the weighted labels of the left context
            for(int distance = 0; distance < leftContext.length; distance++) {
                // If the labels are not of the same dimensionality as the context we should not add them
                if(leftContext[distance].getDimensionality() != _wordContext.length)
                	retval = false;
                tempRL.setContext(weightingScheme.applyLeftWeighting(tempRL, distance+1, leftContext[distance]));
            }
            // Update focusContext with the weighted labels of the right context
            for(int distance = 0; distance < rightContext.length; distance++) {
                // If the labels are not of the same dimensionality as the context we should not add them
                if(rightContext[distance].getDimensionality() != _wordContext.length)
                	retval = false;
                tempRL.setContext(weightingScheme.applyRightWeighting(tempRL, distance+1, rightContext[distance]));
            }
            if(retval)
            	_wordContext = tempRL.getContext();
        }
        _termFrequency++;
    	_validState = true;
        return retval;
    }

    /**
     * Calculate the cosine similarity between this RandomLabel and the given.
     * @param label the RandomLabel that is to be compared with this RandomLabel.
     * @return the 2-norm (Euclidean Distance) between this RandomLabel and
     *         the given.
     **/
    public float cosineSim(RandomLabel label) {
    	SimilarityMeasure sm = new CosineD_row();
    	float[][] matrix = {_wordContext, label.getContext()};
        Float simF = new Float(sm.analyzeMatrix(matrix)[0][1]);
        return simF.isNaN() ? 0 : simF.floatValue();
    }

    /**
     * Compares this RandomLabel with the specified RandomLabel for order on basis of term frequency.
     * Returns a negative integer, zero, or a positive integer as this RandomLabel is less than,
     * equal to, or greater than the specified RandomLabel. This gives descending term frequency
     * when used in a sort.
     *
     * Note: this class has a natural ordering that is inconsistent with equals.
     * 
     * @param label the RandomLabel to be compared.
     * @return a negative integer, zero, or a positive integer as this object is less than,
     *         equal to, or greater than the specified RandomLabel.
     */
    public int compareTo(Object label) throws ClassCastException {
        if(!(label instanceof RandomLabel))
            throw new ClassCastException("A RandomLabel object expected.");

        long termFrequency = ((RandomLabel)label).getTermFrequency();
        long difference = (int)termFrequency - _termFrequency;
        return (difference > 0) ? 1 : (difference < 0) ? -1 : 0;
    }

    /**
     * String representation of the RandomLabel.
     * @return a comma separated string where the first column is the word, the second
     * is meta data, the third is the term frequency, and the fourth is the document 
     * frequency. After this comes a coulmn denoting the number of -+1:s followed by 
     * the positions for that many +1:s and then that many -1:s. Lastly we have a 
     * column denoting the dimensionality of the context vector followed by the 
     * actual context vector itself (until end).
     */
    public String toString() {
        StringBuffer retstring = new StringBuffer();
        retstring.append(_word + "," + _meta_data + "," + _termFrequency + ",");
        retstring.append( + _docFrequency + "," + _poss.length + ",");
        retstring.append(ArrayTools.join(_poss,",") + ",");
        retstring.append(ArrayTools.join(_negs,",") + ",");
        retstring.append(_dimensionality + "," + ArrayTools.join(_wordContext,","));
        return retstring.toString();
    }
}
