/*
 * TagStripper.java
 * Copyright (c) 2004, KTH NADA.
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, see
 * http://opensource.org/licenses/gpl-license.php
 *
 * Martin Hassel, 2004-aug-16
 * http://www.nada.kth.se/~xmartin/
 */
package moj.util;

import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Pattern;

import javax.swing.text.MutableAttributeSet;
import javax.swing.text.html.HTML;
import javax.swing.text.html.HTMLEditorKit;

/**
 * A small utility class that contains methods for stripping tags from 
 * documents. It tries to retain the structure of the original text 
 * by converting line breaks, paragraphs, headings etc to ordinary (hard)
 * single and double line breaks. It also strips comments and converts 
 * entities.
 * 
 * The design for this class is heavily influenced by an example in
 * "Web Client Programming with Java" by Elliotte Rusty Harold.
 *
 * @author  Martin Hassel
 * @version 2006-Dec-08
 */
public class TagStripper {
	// HTML
	public static final Pattern xlt = Pattern.compile("&lt;");
	public static final Pattern xgt = Pattern.compile("&gt;");
	public static final Pattern xamp = Pattern.compile("&amp;");
	public static final Pattern xquot = Pattern.compile("&quot;");
	public static final Pattern xaring = Pattern.compile("&aring;");
	public static final Pattern xauml = Pattern.compile("&auml;");
	public static final Pattern xouml = Pattern.compile("&ouml;");
	public static final Pattern xAring = Pattern.compile("&Aring;");
	public static final Pattern xAuml = Pattern.compile("&Auml;");
	public static final Pattern xOuml = Pattern.compile("&Ouml;");
	// ASCII/Unicode
	public static final Pattern alt = Pattern.compile("<");
	public static final Pattern agt = Pattern.compile(">");
	public static final Pattern aamp = Pattern.compile("&");
	public static final Pattern aquot = Pattern.compile("\"");
	public static final Pattern aring = Pattern.compile("\u00e5");
	public static final Pattern auml = Pattern.compile("\u00e4");
	public static final Pattern ouml = Pattern.compile("\u00f6");
	public static final Pattern Aring = Pattern.compile("\u00c5");
	public static final Pattern Auml = Pattern.compile("\u00c4");
	public static final Pattern Ouml = Pattern.compile("\u00d6");
	// Control code conversion
	private static Map<String,String> cc2name = null;
	private static Map<String,String> name2cc = null;

	/**
	 * Strips HTML tags and comments from the given string and returns a string 
	 * while trying to retain the structure of the original text by converting 
	 * HTML line breaks, paragraphs, headings etc to ordinary (hard) single 
	 * and double line breaks. Also strips comments and converts HTML entities.
	 * 
	 * @param htmltext the string that is to be stripped from HTML
	 * @return the given string stripped from HTML
	 */
	public String stripHTML(String htmltext) {
		// Create a StreamReader you can read the HTML text with.
		return stripHTML(new StringReader(htmltext));
	}

	/**
	 * Strips HTML tags and comments from the given steam and returns a string 
	 * while trying to retain the structure of the original text by converting 
	 * HTML line breaks, paragraphs, headings etc to ordinary (hard) single 
	 * and double line breaks. Also strips comments and converts HTML entities.
	 * 
	 * @param reader the stream that is to be stripped from HTML
	 * @return the given stream stripped from HTML returned as a string
	 */
	public String stripHTML(Reader reader) {
		// StringBuffer to collect the HTML stripped text in.
		final StringBuffer strbuf = new StringBuffer();
		// Begin by retrieving a parser using the ParserGetter class:
		ParserGetter kit = new ParserGetter();
		HTMLEditorKit.Parser parser = kit.getParser();
		// Next, construct an instance of your callback class like this:
		HTMLEditorKit.ParserCallback callback = new LineBreakingTagStripper(strbuf);
		try {
			// Finally, pass the Reader and HTMLEditorKit.ParserCallback to 
			// the HTMLEditorKit.Parser's parse() method, like this:
			parser.parse(reader, callback, false);
		}
		catch(IOException ex) {
			System.err.println(ex); 
		}
		return strbuf.toString();
	}

	/**
	 * Convert markup characters (&lt; &gt; &amp; &quot;) to XML entities 
	 * (&amp;lt; &amp;gt; &amp;amp; &amp;quot;).
	 * @param text <code>String</code> (possibly) containing markup characters
	 * @return <code>String</code> with markup characters converted to XML entities
	 */
	public static String toXMLentities(String text) {
		text = alt.matcher(text).replaceAll("&lt;");
		text = agt.matcher(text).replaceAll("&gt;");
		text = aamp.matcher(text).replaceAll("&amp;");
		text = aquot.matcher(text).replaceAll("&quot;");
//		text = aring.matcher(text).replaceAll("&aring;");
//		text = auml.matcher(text).replaceAll("&auml;");
//		text = ouml.matcher(text).replaceAll("&ouml;");
//		text = Aring.matcher(text).replaceAll("&Aring;");
//		text = Auml.matcher(text).replaceAll("&Auml;");
//		text = Ouml.matcher(text).replaceAll("&Ouml;");
		return text;
	}

	/**
	 * Convert XML entities (&amp;lt; &amp;gt; &amp;amp; &amp;quot;) to markup 
	 * characters (&lt; &gt; &amp; &quot;).
	 * @param text <code>String</code> (possibly) containing XML entities
	 * @return <code>String</code> with XML entities converted to markup characters
	 */
	public static String fromXMLentities(String text) {
		text = xlt.matcher(text).replaceAll("<");
		text = xgt.matcher(text).replaceAll(">");
		text = xamp.matcher(text).replaceAll("&");
		text = xquot.matcher(text).replaceAll("\"");
//		text = xaring.matcher(text).replaceAll("\u00e5");
//		text = xauml.matcher(text).replaceAll("\u00e4");
//		text = xouml.matcher(text).replaceAll("\u00f6");
//		text = xAring.matcher(text).replaceAll("\u00c5");
//		text = xAuml.matcher(text).replaceAll("\u00c4");
//		text = xOuml.matcher(text).replaceAll("\u00d6");
		return text;
	}

	/**
	 * Replaces all C0/C1 control codes in <code>text</code> with whitespace.
	 * @param text <code>String</code> to remove C0/C1 control codes from
	 * @return <code>String</code> with C0/C1 control removed
	 */
	public static String removeControlCodes(String text) {
		if(cc2name == null)
			populatecc2name();
		for(String cc : cc2name.keySet())
			text = Pattern.compile(cc).matcher(text).replaceAll(" ");
		return text;
	}
	
	/**
	 * Replaces all C0/C1 control codes in <code>text</code> with its abbreviated name.
	 * @param text <code>String</code> to replace C0/C1 control codes in
	 * @return <code>String</code> with C0/C1 control replaced
	 */
	public static String fromControlCodes(String text) {
		if(cc2name == null)
			populatecc2name();
		for(String cc : cc2name.keySet())
			text = Pattern.compile(cc).matcher(text).replaceAll(cc2name.get(cc));
		return text;
	}
	
	// Results unpredictable, private for the time being
	@SuppressWarnings("unused")
	private static String toControlCodes(String text) {
		if(cc2name == null)
			populatename2cc();
		for(String name : name2cc.keySet())
			text = Pattern.compile(name).matcher(text).replaceAll(name2cc.get(name));
		return text;
	}
	
	// Mappning from C0/C1 control codes to short identifier 
	private static void populatecc2name() {
		cc2name = new HashMap<String,String>();
		cc2name.put("\u0000","NUL");
		cc2name.put("\u0001","SOH");
		cc2name.put("\u0002","STX");
		cc2name.put("\u0003","ETX");
		cc2name.put("\u0004","EOT");
		cc2name.put("\u0005","ENQ");
		cc2name.put("\u0006","ACK");
		cc2name.put("\u0007","BEL");
		cc2name.put("\u0008","BS");
		cc2name.put("\u0009","HT");
		/*cc2name.put("\u000A","LF");*/
		cc2name.put("\u000B","VT");
		cc2name.put("\u000C","FF");
		/*cc2name.put("\u000D","CR");*/
		cc2name.put("\u000E","SO");
		cc2name.put("\u000F","SI");
		cc2name.put("\u0010","DLE");
		cc2name.put("\u0011","DC1");
		cc2name.put("\u0012","DC2");
		cc2name.put("\u0013","DC3");
		cc2name.put("\u0014","DC4");
		cc2name.put("\u0015","NAK");
		cc2name.put("\u0016","SYN");
		cc2name.put("\u0017","ETB");
		cc2name.put("\u0018","CAN");
		cc2name.put("\u0019","EM");
		cc2name.put("\u001A","SUB");
		cc2name.put("\u001B","ESC");
		cc2name.put("\u001C","FS");
		cc2name.put("\u001D","GS");
		cc2name.put("\u001E","RS");
		cc2name.put("\u001F","US");
		cc2name.put("\u007F","DEL");
		cc2name.put("\u0080","PAD");
		cc2name.put("\u0081","HOP");
		cc2name.put("\u0082","BPH");
		cc2name.put("\u0083","NBH");
		cc2name.put("\u0084","IND");
		cc2name.put("\u0085","NEL");
		cc2name.put("\u0086","SSA");
		cc2name.put("\u0087","ESA");
		cc2name.put("\u0088","HTS");
		cc2name.put("\u0089","HTJ");
		cc2name.put("\u008A","VTS");
		cc2name.put("\u008B","PLD");
		cc2name.put("\u008C","PLU");
		cc2name.put("\u008D","RI");
		cc2name.put("\u008E","SS2");
		cc2name.put("\u008F","SS3");
		cc2name.put("\u0090","DCS");
		cc2name.put("\u0091","PU1");
		cc2name.put("\u0092","PU2");
		cc2name.put("\u0093","STS");
		cc2name.put("\u0094","CCH");
		cc2name.put("\u0095","MW");
		cc2name.put("\u0096","SPA");
		cc2name.put("\u0097","EPA");
		cc2name.put("\u0098","SOS");
		cc2name.put("\u0099","SGC");
		cc2name.put("\u009A","SCI");
		cc2name.put("\u009B","CSI");
		cc2name.put("\u009C","ST");
		cc2name.put("\u009D","OSC");
		cc2name.put("\u009E","PM");
		cc2name.put("\u009F","APC");
	}

	// Mappning from short identifier to C0/C1 control codes
	private static void populatename2cc() {
		name2cc = new HashMap<String,String>();
		name2cc.put("NUL","\u0000");
		name2cc.put("SOH","\u0001");
		name2cc.put("STX","\u0002");
		name2cc.put("ETX","\u0003");
		name2cc.put("EOT","\u0004");
		name2cc.put("ENQ","\u0005");
		name2cc.put("ACK","\u0006");
		name2cc.put("BEL","\u0007");
		name2cc.put("BS","\u0008");
		name2cc.put("HT","\u0009");
		/*name2cc.put("LF","\u000A");*/
		name2cc.put("VT","\u000B");
		name2cc.put("FF","\u000C");
		/*name2cc.put("CR","\u000D");*/
		name2cc.put("SO","\u000E");
		name2cc.put("SI","\u000F");
		name2cc.put("DLE","\u0010");
		name2cc.put("DC1","\u0011");
		name2cc.put("DC2","\u0012");
		name2cc.put("DC3","\u0013");
		name2cc.put("DC4","\u0014");
		name2cc.put("NAK","\u0015");
		name2cc.put("SYN","\u0016");
		name2cc.put("ETB","\u0017");
		name2cc.put("CAN","\u0018");
		name2cc.put("EM","\u0019");
		name2cc.put("SUB","\u001A");
		name2cc.put("ESC","\u001B");
		name2cc.put("FS","\u001C");
		name2cc.put("GS","\u001D");
		name2cc.put("RS","\u001E");
		name2cc.put("US","\u001F");
		name2cc.put("DEL","\u007F");
		name2cc.put("PAD","\u0080");
		name2cc.put("HOP","\u0081");
		name2cc.put("BPH","\u0082");
		name2cc.put("NBH","\u0083");
		name2cc.put("IND","\u0084");
		name2cc.put("NEL","\u0085");
		name2cc.put("SSA","\u0086");
		name2cc.put("ESA","\u0087");
		name2cc.put("HTS","\u0088");
		name2cc.put("HTJ","\u0089");
		name2cc.put("VTS","\u008A");
		name2cc.put("PLD","\u008B");
		name2cc.put("PLU","\u008C");
		name2cc.put("RI","\u008D");
		name2cc.put("SS2","\u008E");
		name2cc.put("SS3","\u008F");
		name2cc.put("DCS","\u0090");
		name2cc.put("PU1","\u0091");
		name2cc.put("PU2","\u0092");
		name2cc.put("STS","\u0093");
		name2cc.put("CCH","\u0094");
		name2cc.put("MW","\u0095");
		name2cc.put("SPA","\u0096");
		name2cc.put("EPA","\u0097");
		name2cc.put("SOS","\u0098");
		name2cc.put("SGC","\u0099");
		name2cc.put("SCI","\u009A");
		name2cc.put("CSI","\u009B");
		name2cc.put("ST","\u009C");
		name2cc.put("OSC","\u009D");
		name2cc.put("PM","\u009E");
		name2cc.put("APC","\u009F");
	}

	private class ParserGetter extends HTMLEditorKit {
		private static final long serialVersionUID = 2883160159008600523L;
		// Purely to make this method public
		public HTMLEditorKit.Parser getParser() {
			return super.getParser();
		}
	}
	
	private class LineBreakingTagStripper extends HTMLEditorKit.ParserCallback {
		private StringBuffer _strbuf;
		private String _lineSeparator;
		
		public LineBreakingTagStripper(StringBuffer out) {
			this(out, System.getProperty("line.separator", "\r\n")); 
		}  
		
		public LineBreakingTagStripper(StringBuffer out, String lineSeparator) {
			_strbuf = out;
			_lineSeparator = lineSeparator;
		}  
		
		public void handleText(char[] text, int position) {
			_strbuf.append(text);
		}
		
		public void handleEndTag(HTML.Tag tag, int position) {
			if(tag.isBlock()) {
				_strbuf.append(_lineSeparator);
				_strbuf.append(_lineSeparator);
			} else if(tag.breaksFlow()) {
				_strbuf.append(_lineSeparator);
			}
		}
		
		public void handleSimpleTag(HTML.Tag tag, MutableAttributeSet attributes, int position) {
			if(tag.isBlock()) {
				_strbuf.append(_lineSeparator);
				_strbuf.append(_lineSeparator);
			} else if (tag.breaksFlow()) {
				_strbuf.append(_lineSeparator);
			} else {
				_strbuf.append(' '); 
			}
		}
	}
}
